/* This module implements short-term national targets to 2020 and 2050 as pledged following the COP15 Copenhagen deal.
 * It precedes "shares" module which fills in the national targets after this period
 */
package jcm.mod.regemit;

import java.util.*;
import jcm.core.*;
import jcm.core.cur.*;
import jcm.core.par.param;
import jcm.core.ob.module;
import jcm.core.reg.*;
import jcm.core.ob.loopcalc;
import jcm.mod.socio.popgdp;
import static jcm.core.complexity.*;
import static jcm.core.report.*;
import static jcm.gui.gen.colfont.*;
import static jcm.mod.regemit.pledges.kyoto_opts.*;
import static jcm.mod.regemit.pledges.gdpopts.*;
import static jcm.mod.regemit.pledges.pledgelevel.*;
import static jcm.mod.regemit.pledges.targettype.*;

public class pledges extends module {
    //============== PARAMS ==========
    public  enum pledgelevel {
        unconditional, conditional, mixed
    }

    public enum kyoto_opts {
        real_emissions, Kyoto_targets_excUSA, Kyoto_targets_excUSA_prolonged, Kyoto_targets_incUSA
    };


    public enum gdpopts {
        GDP_PPP, GDP_MER
    };


    public param usepledges = new param("apply pledges", true);
    public param beyond2020 = new param("apply beyond 2020", false); 
    public param EU2050 = new param("EU targets 2050", false, expert);
    public param<pledgelevel> level = new param("pledge level", pledgelevel.values(), conditional);
    public param<kyoto_opts> kyoto = new param("Kyoto Protocol", kyoto_opts.values(), Kyoto_targets_excUSA_prolonged);
    public param sephotair = new param("Separate AAUsurplus", true);
    public param AAUfrac = new param("fraction surplus AAUs used", 50, 0, 100, "percent");
    public param<gdpopts> choosegdp = new param("gdp for intensity", gdpopts.values(), GDP_PPP, expert);
  

// ============ VARIABLES=========
    public Map<region, Integer> endtarget = new HashMap(50); //stores end year of target
    public Set<region> notarget = new LinkedHashSet(50); //countries with no target
    Set<region> EU = new LinkedHashSet(27);
    public Map<region, List<target>> targets = new LinkedHashMap();
    emitquota eq;
    emitbase eb;
    boolean fixedEU=false;
     curveset efb, efq, hfos, gdp_mer, gdp_ppp;

       //==================== INTERACTIONS ===========
    //
    public void initsetup() {
        eq = gm(emitquota.class);
        eb = gm(emitbase.class);
        follows(gm(emitbase.class));
        follows(gm(AviaShipEmit.class).emitbase);
        setaffectedby(gm(shares.class)); //not efficient - in principle pledges part shouldn't change unless shares over-rides pledges - but it can
        EU=regman.allreg.findreg("EU27").subreg(regman.nations);
        settargets();
    }

    public void setinteractions() {
        setaffectedby(level, usepledges.istrue());
        setaffectedby(kyoto, usepledges.istrue());
        setaffectedby(choosegdp, usepledges.istrue());
        setaffectedby(beyond2020, usepledges.istrue());
        setaffectedby(sephotair, (usepledges.istrue() || !(kyoto.chosen==real_emissions)));
        setaffectedby(AAUfrac,  (usepledges.istrue() || !(kyoto.chosen==real_emissions)) && sephotair.istrue());
        setaffectedby(EU2050, beyond2020.istrue() && usepledges.istrue());
    }
     //**************** LOOP METHODS *********
    public void precalc() {

        efb =eb.emitfosbase_nat;
        efq= eq.emitfos_nat;
        gdp_mer = gm(popgdp.class).gdp_mer_nat;
        gdp_ppp = gm(popgdp.class).gdp_ppp_nat;
        hfos = histemitdata.fosCO2; //beware in kilo-ton not mega-ton!
       if (!fixedEU) fixEU(); //normalise total to 20% low and 30% high
        for (region r : EU) endtarget.put(r,  EU2050.istrue() ? 2050 : 2020);
        applytargets();
        /*apply targets can run in precalc only presuming no dependence on other modules within calcstep loop?
        eg popgdp - but that also works entirely in precalc
        and emitbase - also works in precalc except totals and bunker
        */
        fillhistory();  
        /* fill history is for countries without any target
       in principle shouldn't need to do fill history more than once - check if can do in initsetup - but beware changing start year in shares etc.
        also might want option to remove all without targets - should be an option in shares?
        */
         }


    //==================== TARGETS  ===========
    public enum targettype {
        absolute, intensity, baseline;
    }

    /*use a class to group target info together (not an enum, because targets may need to be user-adjustable)
     *may need several targets for different years, for each country
     *may have targets referring to a group of countries
     */
    public class target {
        public region reg;
        public int baseyear,  targetyear;
        public targettype type;
        public float low,  high,  likely; 
        target(region r, targettype t, int b, int y, float l, float h, float v) {
            reg = r;
            type = t;
            baseyear = b;
            targetyear = y;
            low = l;
            high = h;
            likely = v;
        }

    } // class target



    //========== SET TARGETS ==========
    //note: set targets for groups, followed by nations within them, the latter will over-ride group target
    void settargets() {
        st("USA", 1990, 2010, -7, 2005, 2020, -17, 2025, -30, 2030, -42, 2050, -83);
        st("Australia", 1990, 2010, +8, 2000, 2020, -5,  -25, -15); // note this implies that -15 stored as the "most likely" - should check which is actually most likely??
        st("New Zealand", 1990, 2010, 0, 2020, -10, -20); //check if recent change ?
        st("Canada", 1990, 2010, -6, 2005, 2020, -17); //before copenhagen was -20 wrt 2006
        st("Japan", 1990, 2010, -6, 2020, -15, -25); //earlier in 2009 was -15
        st("Russia", 1990, 2010, 0, 2020, -15, -25); //adjusted - lower pledge was -22 pre Cop?
        st("Ukraine", 1990, 2010, 0, 2020, -15, -25); //as same same as russia? -CHECK!
        st("Belarus", 1990, 2010, -8, 2020, -5, -10); //new pledge jan 2010
        st("Kazakhstan", 1992, 2020, -15); //new pledge jan 2010

        //figures for EU15 for 1990 are from EU burden sharing agreement, for the 10 eastern countries they are from Kyoto protocol - note 5 have different base years
        //figures for 2020 below are taken from  + EU effort-sharing agreement for -20%apr09 http://eur-lex.europa.eu/LexUriServ/LexUriServ.do?uri=OJ:L:2009:140:0136:0148:EN:PDF ANNEX II
        //note normalisation to 20% total reduction in fixEU method
        st("Austria", 1990, 2010, -13, 2005, 2020, -16);
        st("Belgium", 1990, 2010, -7.5, 2005, 2020, -15);
        st("Bulgaria", 1988, 2010, -8, 2005, 2020, +20);
        st("Cyprus", 2005, 2020, -5);
        st("Czech Republic", 1990, 2010, -8, 2005, 2020, +9);
        st("Denmark", 1990, 2010, -21, 2005, 2020, -20);
        st("Estonia", 1990, 2010, -8, 2005, 2020, +11);
        st("Finland", 1990, 2010, 0, 2005, 2020, -16);
        st("France", 1990, 2010, 0, 2005, 2020, -14);
        st("Germany", 1990, 2010, -21, 2005, 2020, -14);
        st("Greece", 1990, 2010, +25, 2005, 2020, -4);
        st("Hungary", 1986, 2010, -6, 2005, 2020, +10); //technically, base year for Kyoto is average of 85-87
        st("Ireland", 1990, 2010, +13, 2005, 2020, -20);
        st("Italy", 1990, 2010, -6.5, 2005, 2020, -13);
        st("Latvia", 1990, 2010, -8, 2005, 2020, +17);
        st("Lithuania", 1990, 2010, -8, 2005, 2020, +15);
        st("Luxembourg", 1990, 2010, -28, 2005, 2020, -20);
        st("Malta", 2005, 2020, +5);
        st("Netherlands", 1990, 2010, -6, 2005, 2020, -16);
        st("Poland", 1988, 2010, -6, 2005, 2020, +14);
        st("Portugal", 1990, 2010, +27, 2005, 2020, +1);
        st("Romania", 1989, 2010, -8, 2005, 2020, +19);
        st("Slovenia", 1986, 2010, -8, 2005, 2020, +4);
        st("Slovakia", 1990, 2010, -8, 2005, 2020, +13);
        st("Spain", 1990, 2010, +15, 2005, 2020, -10);
        st("Sweden", 1990, 2010, +4, 2005, 2020, -17);
        st("United Kingdom", 1990, 2010, -12.5, 2005, 2020, -16); //using EU effort-sharing for -20%
//st("United Kingdom", 1990,  2010, -12.5, 2020, -34, -42, 2050, -80); //uk domestic target -34% by 2020, -80% by 2050

        st("Croatia", 1990, 2010, -5, 2020, -5); //CHECK IF 2020 NUMBER CORRECT?
        st("Switzerland", 1990, 2010, -8, 2020, -20, -30); //for 2020, apply same as EU temporarily, as leaving nothing causes strange results (CHECK!)
        st("Norway", 1990, 2010, +1, 2020, -30, -40); // -40 is original annoucement, the -30 low end was added jan 2010 in response to low ambition of others
        st("Iceland", 1990, 2010, +10, 2020, -30);
 
        st("Brazil", baseline, 2020, -36.1, -38.9); //Brazil's pledge is defined according to specific reductions in each sector, the baseline figure is derived from that
        //st("Mexico", 2002, 2012, -8, 2050, -50); //check Special Programme on CC
        st("Mexico", baseline, 2020, -30); //For Mexico, the original plan was quite complex, and specified in GtC for each sector rather than %, hard to interpret simply.
        st("Korea", baseline, 2005, 2020, -30); //pledge is related to baseline - note range is 21-30
        //st("Korea", 2005, 2020, -4); //widely cited intrepretation of above
        st("South Africa", baseline, 2020, -34, 2025, -42);
        st("Indonesia", baseline, 2020, -26); //, -41); //Indonesia announced a target of -41 before Copenhagen, dependent on ambition and  financing  from other countries
        st("Singapore", baseline, 2020, -16); //see Sustainable Singapore Blueprint in April 2009

        st("China", intensity, 2005, 2020, -40, -45);
        st("India", intensity, 2005, 2020, -20, -25); //, 2030 , -37);  pre-cop15 India said -24 by 2020, -37 by 2030, but latter was not retained in cop-accord


        notarget.addAll(regman.nations.reg);
        notarget.removeAll(targets.keySet());

    } //end settargets

    //st("EU27", 1990,  2010, -8, 2020, -20, -30, 2050, -80, -90);
// in principle this should just extend the above targets to 2050 -



    void st(String s, Number... i) {
        st(s, targettype.absolute, i);
    }

    void st(String regname, targettype type, Number... i) {
        region r = regman.allreg.findreg(regname);
        if (r == null) {
            deb("Targets error: no region " + regname);
            return;
        }
        List<target> lt = (!targets.containsKey(r) ? new ArrayList(3) : targets.get(r));

        int baseyear = 0, targetyear = 0;
        boolean gotyear = false;
        int count = 0;
        float low = 0, high = 0, value = 0;
        for (Number j : i) {
            if (j.intValue() > 1000) /*it's a year*/ {
                if (count > 0) /*start next target, save old one */ {
                    lt.add(new target(r, type, baseyear, targetyear, low, high, value));
                    count = 0;
                }
                if (gotyear) baseyear = targetyear;
                targetyear = j.intValue();
                gotyear = true;
            } else {
                gotyear = false;
                count++;
                if (count == 1) {
                    low = j.floatValue();
                    high = low;
                    value = low;
                }
                if (count == 2) {
                    high = j.floatValue();
                    value = (low + high) / 2;
                }
                if (count == 3) {
                    value = j.floatValue();
                }
            }
        }
        lt.add(new target(r, type, baseyear, targetyear, low, high, value)); //save last one
        endtarget.put(r, targetyear);

        if (regman.nations.contains(r)) {
            targets.put(r, lt);
            return;
        }
        //if it's a region-set apply to all nations within region
        for (region sr : r.subreg(regman.nations)) targets.put(sr, lt);

    }



    public void applytargets() {
       //note efb and efq are in megaton carbon, whereas history and target in kiloton - hence the 1000f factors
        //note changed jun2011 to use gdp_nat - previously was regional = bug if regionset not separating india and china?
        curveset gdp = (choosegdp.chosen == GDP_PPP) ? gdp_ppp: gdp_mer;

        for (int y=1990; y<=2020; y++) { efq.set("AAUsurplus", y, 0); }

        for (region r : targets.keySet()) {
            //if (!regman.nations.contains(r)) { deb("Error: can only apply targets to nations, not to "+r.name); return; }
            int year = 1990;
            float base, target, bau, tarlin=0;

            //go through targets (earlier first) without resetting  year
            for (target t : targets.get(r)) {
                base = hfos.get(r, t.baseyear) / 1000f; //note the assumption is that 2005 targets  apply to actual emissions not to kyoto levels
                target = ((100f + 
                        (level.chosen==conditional? t.high : level.chosen==unconditional ? t.low : t.likely)
                        ) / 100f) * (t.type == baseline ? efb.get(r, t.targetyear) : base * (t.type == intensity ? gdp.get(r, t.targetyear) / gdp.get(r, t.baseyear) : 1f));

                //start by setting to real emissions up to 2008 (if not done already)
                while (year <= 2007) {
                    efq.set(r, year, hfos.get(r, year) / 1000f);
                    year++;
                }

                //kyoto targets 2008-12
                if (t.targetyear == 2010 && kyoto.chosen!=real_emissions && (kyoto.chosen==Kyoto_targets_incUSA || !r.name.equals("USA"))) {
                        while (year <= 2012) {
                            bau=hfos.get(r, year) / 1000f;
                            if (target>bau && sephotair.istrue()&& year<=2020) {
                                efq.set("AAUsurplus", year, efq.get("AAUsurplus", year)+(target-bau)*(float)AAUfrac.getval()/100f);
                                efq.set(r, year, bau);
                            }
                            else  efq.set(r, year, target);
                             year++;
                        }
                        tarlin = (kyoto.chosen==Kyoto_targets_excUSA_prolonged) ? target : hfos.get(r, 2012) / 1000f;
                }
                //else continue with real emissions up to 2012
                else while (year <= 2012) {
                    efq.set(r, year, hfos.get(r, year) / 1000f);
                    if (year==2012) tarlin= hfos.get(r, year) / 1000f;
                    year++;
                }

                //apply linear approach towards 2020 target
                while (year <= t.targetyear) {
                    bau=efb.get(r, year);
                    tarlin=(tarlin * (t.targetyear - year) + target) / (1 + t.targetyear - year);
                    if (tarlin>bau && sephotair.istrue() && year<=2020) {
                        //deb("year="+year+" region="+r+ "surp=" + (tarlin-bau));
                        efq.set("AAUsurplus", year,  efq.get("AAUsurplus", year)+(tarlin-bau)*(float)AAUfrac.getval()/100f);
                        efq.set(r, year, bau);
                    }
                    else efq.set(r, year, tarlin);
                    //deb("year="+year+" region="+r+ "history="+history+" t.baseyear="+t.baseyear);
                   // deb("target for "+r.name+" in "+year+" is "+eq.fosCO2_nat.get(r, year));
                    year++;
                }
            } //target
        //while (year<=2050) {fosCO2_nat.set(r, year, curve.dud); year++; } // old code to fill in blanks for reset after remove target??
        } //region
         for (int y=1990; y<=2020; y++) { //copy from nat to region etc
             float surplus= efq.get("AAUsurplus", y);
             eq.emitfos.set("AAUsurplus", y, surplus);
             eq.emitfos_unitsCO2.set("AAUsurplus", y, surplus*44f/12f); 
             eq.emitequiv.set("AAUsurplus", y, surplus*44f/12f); 
             eq.emitequiv_nat.set("AAUsurplus", y, surplus*44f/12f); 
         }
    }

// fix EU, scaling as a group - necessary because burden sharing agreement is only for domestic part not for large industry covered by EU-ETS, so original numbers don't add to -20%
    void fixEU() {
    float tot2020=0, totbase=0;
        for (region r : EU) {
            List<target> tt=targets.get(r);
            target t=tt.get(tt.size()-1); // the last target - as not all have a Kyoto target
            tot2020+=hfos.get(r,t.baseyear) / 1000f *(100f+t.low)/100f; //note baseyear is 2005
            totbase+=hfos.get(r,1990) /1000f ; //but group target is wrt 1990...
        }
     float faclow20=0.8f*totbase/tot2020, fachigh30=0.7f*totbase/tot2020;
     float faclow80=0.2f*totbase/tot2020, fachigh95=0.05f*totbase/tot2020;
            //deb("faclow="+faclow+" fachigh="+fachigh);
        for (region r : EU) {
            List<target> tt=targets.get(r);
            target t=tt.get(tt.size()-1); // the last target - as not all have a Kyoto target
            //first make a 2050 target (with same relative shares as 2020)
            float t2low=(100f+t.low)*faclow80 - 100f, t2high=(100f + t.high)*fachigh95 -100f, t2likely=(t2low+t2high)/2f;
            //now correct the 2020 target
            t.low=(100f+t.low)*faclow20 - 100f; t.high=(100f + t.high)*fachigh30 -100f; t.likely=(t.low+t.high)/2f;
            //deb("targets for "+r+" low: "+t.low+" likely: "+t.likely+" high: "+t.high);
            tt.add(new target(r, absolute, 1990, 2050, t2low, t2high, t2likely));
            endtarget.put(r, 2050);
        }
        fixedEU=true;
    }

    public void fillhistory() {
        for (region r : notarget) {
            for (int y = 1990; y <=hfos.ey; y++)   eq.emitfos_nat.set(r, y, hfos.get(r, y) / 1000f);
            for (int y = hfos.ey + 1; year <=2050; y++) eq.emitfos_nat.set(r, y, curve.dud);
        }
        for (int y = 1990; y <=  hfos.ey; y++) eq.emitfos_nat.set("bunker", y, eb.emitfosbase.get("bunker", y));
        for (int y = 1990; y <=  hfos.ey; y++) eq.emitfos_nat.calctot(y);
    }

} //end class
//========================



